from iam_syncr.errors import SyncrError, BadAmazon

from boto.iam.connection import IAMConnection
from contextlib import contextmanager
import logging
import urllib
import json
import boto
import sys

log = logging.getLogger("iam_syncr.amazon")

class LeaveAlone(object):
    """Used to differentiate between None and not specified in a call signature"""

class Amazon(object):

    # Account info that is overridden by __init__
    accounts = None
    account_id = None
    account_name = None

    @classmethod
    def set_boto_useragent(self, app_name, version):
        """Put this app in the useragent used by boto"""
        __import__("boto")
        useragent = sys.modules["boto.connection"].UserAgent
        if app_name not in useragent:
            sys.modules["boto.connection"].UserAgent = "{0} {1}/{2}".format(useragent, app_name, version)

    def __init__(self, account_id, account_name, accounts):
        self.changes = False
        self.accounts = accounts
        self.account_id = account_id
        self.account_name = account_name

    def setup(self):
        """Make sure our current credentials are for this account and set self.connection"""
        try:
            connection = IAMConnection()
        except boto.exception.NoAuthHandlerFound:
            raise SyncrError("Export AWS_ACCESS_KEY_ID and AWS_SECRET_ACCESS_KEY before running this script (your aws credentials)")

        try:
            result = connection.list_roles()
        except boto.exception.BotoServerError as error:
            if error.status == 403:
                raise SyncrError("Your credentials aren't allowed to look at iam :(")
            else:
                raise

        roles = result["list_roles_response"]["list_roles_result"]["roles"]
        if not roles:
            raise SyncrError("There are no roles in your account, I can't figure out the account id")

        amazon_account_id = roles[0]['arn'].split(":")[4]
        if str(self.account_id) != str(amazon_account_id):
            raise SyncrError("Please use credentials for the right account", expect=self.account_id, got=amazon_account_id)

        # If reached this far, the credentials belong to the correct account :)
        self.connection = connection
        return connection

    @contextmanager
    def catch_boto_400(self, message, heading=None, document=None, **info):
        """Turn a BotoServerError 400 into a BadAmazon"""
        try:
            yield
        except boto.exception.BotoServerError as error:
            if error.status == 400:
                if heading or document:
                    print("=" * 80)
                    if heading:
                        print(heading)
                    print(document)
                    print("=" * 80)
                raise BadAmazon(message, error_code=error.code, error_message=error.message, **info)
            else:
                raise

    def split_role_name(self, name):
        """Split a role name into it's (name, path)"""
        split = name.split('/')
        role_name = split[-1]
        path = '/'.join(split[0:-1]) or None
        if path and not path.startswith('/'):
            path = "/{0}".format(path)
        if path and not path.endswith("/"):
            path = "{0}/".format(path)
        return role_name, path

    def role_info(self, name):
        """Return what amazon knows about this role"""
        try:
            role_name, _ = self.split_role_name(name)
            return self.connection.get_role(role_name)["get_role_response"]["get_role_result"]
        except boto.exception.BotoServerError as error:
            if error.status == 404:
                return False
            raise

    def has_role(self, name):
        """Return whether amazon has info about this role"""
        return bool(self.role_info(name))

    def info_for_profile(self, name):
        """Return what roles are attached to this profile if it exists"""
        role_name, _ = self.split_role_name(name)
        with self.catch_boto_400("Couldn't list instance profiles associated with a role", role=role_name):
            result = self.connection.list_instance_profiles_for_role(role_name)
        profiles = result["list_instance_profiles_for_role_response"]["list_instance_profiles_for_role_result"]["instance_profiles"]

        existing_profile = None
        for profile in profiles:
            if profile["instance_profile_name"] == name:
                existing_profile = profile
                break

        if existing_profile:
            if "member" in existing_profile["roles"]:
                return [existing_profile["roles"]["member"]["role_name"]]
            else:
                return []

    def make_instance_profile(self, name):
        """Make an instance profile with this name containing this role"""
        role_name, _ = self.split_role_name(name)
        existing_roles_in_profile = self.info_for_profile(role_name)
        if existing_roles_in_profile is None:
            try:
                log.info("Making instance profile\tname=%s", role_name)
                with self.catch_boto_400("Couldn't create instance profile", instance_profile=role_name):
                    self.connection.create_instance_profile(role_name)
                self.changes = True
            except boto.exception.BotoServerError as error:
                if error.status == 409 and error.code == "EntityAlreadyExists":
                    # I'd rather ignore this conflict, than list all the instance_profiles
                    # Basically, the instance exists but isn't associated with the role
                    pass
                else:
                    raise

        if existing_roles_in_profile and any(rl != role_name for rl in existing_roles_in_profile):
            for role in [rl for rl in existing_roles_in_profile if rl != role_name]:
                log.info("Removing role from existing instance_profile\tprofile=%s\trole=%s", role_name, role)
                with self.catch_boto_400("Couldn't remove role from an instance profile", profile=role_name, role=role):
                    self.connection.remove_role_from_instance_profile(role_name, role)
                self.changes = True

        if not existing_roles_in_profile or not any(rl == role_name for rl in existing_roles_in_profile):
            log.info("Adding role to an instance_profile\tprofile=%s\trole=%s", role_name, role_name)
            with self.catch_boto_400("Couldn't add role to an instance profile", role=role_name, instance_profile=role_name):
                self.connection.add_role_to_instance_profile(role_name, role_name)
            self.changes = True

    def create_role(self, name, trust_document, policies=None):
        """Create a role"""
        role_name, role_path = self.split_role_name(name)
        with self.catch_boto_400("Couldn't create role", "{0} trust document".format(name), trust_document, role=name):
            log.info("Creating a new role\trole=%s", name)
            self.connection.create_role(role_name, assume_role_policy_document=trust_document, path=role_path)
            self.changes = True

        # And add our permissions
        if policies:
            for policy_name, document in policies.items():
                if document:
                    with self.catch_boto_400("Couldn't add policy", "{0} - {1} Permission document".format(role_name, policy_name), document, role=role_name, policy_name=policy_name):
                        self.connection.put_role_policy(role_name, policy_name, document)
                        self.changes = True

    def compare_trust_document(self, role_info, trust_document):
        """Say whether the provided trust document is the same as the one in the role_info"""
        if not role_info or not role_info.get("role", {}).get("assume_role_policy_document"):
            return False

        unquoted = urllib.unquote(role_info["role"]["assume_role_policy_document"])
        return self.compare_two_documents(unquoted, trust_document)

    def compare_two_documents(self, doc1, doc2):
        """Compare two documents by converting them into json objects and back to strings and compare"""
        try:
            first = json.dumps(json.loads(doc1), indent=2, sort_keys=True).strip()
        except (ValueError, TypeError):
            return False

        try:
            second = json.dumps(json.loads(doc2), indent=2, sort_keys=True).strip()
        except (ValueError, TypeError):
            return False

        return first == second

    def modify_role(self, role_info, name, trust_document, policies=LeaveAlone):
        """Modify a role"""
        role_name, _ = self.split_role_name(name)
        if trust_document and not self.compare_trust_document(role_info, trust_document):
            with self.catch_boto_400("Couldn't modify trust document", "{0} assume document".format(role_name), trust_document, role=role_name):
                log.info("Modifying trust document\trole=%s", role_name)
                self.connection.update_assume_role_policy(role_name, trust_document)
                self.changes = True

        if policies is LeaveAlone:
            return
        elif policies is None:
            policies = {}

        unknown = []
        with self.catch_boto_400("Couldn't get policies for a role", role=role_name):
            current_policies = self.current_role_policies(role_name, comparing=[pn for pn in policies])
        unknown = [key for key in current_policies if key not in policies]

        if unknown:
            log.info("Role has unknown policies that will be disassociated\trole=%s\tunknown=%s", role_name, unknown)
            for policy in unknown:
                with self.catch_boto_400("Couldn't delete a policy from a role", policy=policy, role=role_name):
                    self.connection.delete_role_policy(role_name, policy)
                self.changes = True

        for policy, document in policies.items():
            if not document:
                if policy in current_policies:
                    log.info("Removing policy\trole=%s\tpolicy=%s", role_name, policy)
                    with self.catch_boto_400("Couldn't delete a policy from a role", policy=policy, role=role_name):
                        self.connection.delete_role_policy(role_name, policy)
                    self.changes = True
            else:
                needed = False
                if policy in current_policies:
                    if not self.compare_two_documents(current_policies.get(policy), document):
                        log.info("Overriding existing policy\trole=%s\tpolicy=%s", role_name, policy)
                        needed = True
                else:
                    log.info("Adding policy to existing role\trole=%s\tpolicy=%s", role_name, policy)
                    needed = True

                if needed:
                    with self.catch_boto_400("Couldn't add policy document", "{0} - {1} policy document".format(role_name, policy), document, role=role_name, policy=policy):
                        self.connection.put_role_policy(role_name, policy, document)
                        log.debug(policy)
                        log.debug(document)
                        log.debug('------')

    def current_role_policies(self, name, comparing):
        """Get the current policies for some role"""
        role_name, _ = self.split_role_name(name)
        with self.catch_boto_400("Couldn't get policies for a role", role=name):
            policies = self.connection.list_role_policies(role_name)["list_role_policies_response"]["list_role_policies_result"]["policy_names"]

        found = {}
        for policy in policies:
            document = None
            if policy in comparing:
                with self.catch_boto_400("Couldn't get policy document for some policy", policy=policy, role=name):
                    doc = self.connection.get_role_policy(role_name, policy)["get_role_policy_response"]["get_role_policy_result"]["policy_document"]
                document = json.dumps(json.loads(urllib.unquote(doc)), indent=2).strip()
            found[policy] = document

        return found

    def remove_role(self, name):
        """Remove the role if it exists"""
        role_name, _ = self.split_role_name(name)
        if self.has_role(role_name):
            log.info("Deleting role\trole=%s", role_name)
            with self.catch_boto_400("Couldn't delete a role", role=role_name):
                self.connection.delete_role(role_name)
            self.changes = True
        else:
            log.info("Role already deleted\trole=%s", role_name)




from iam_syncr.errors import InvalidDocument, BadRole, BadPolicy
from iam_syncr.helpers import listify, listified, as_list

import logging
import json

log = logging.getLogger("iam_syncr.roles")

class RoleRemoval(object):
    def __init__(self, name, amazon):
        self.name = name
        self.amazon = amazon

    def setup(self):
        """Make sure our name is a string"""
        if not isinstance(self.name, basestring):
            raise BadRole("Told to remove a role, but not specified as a string", name=self.name, found_type=type(self.name))

    def resolve(self):
        """Remove the role"""
        self.amazon.remove_role(self.name)

class Role(object):
    def __init__(self, name, definition, amazon):
        self.name = name
        self.amazon = amazon
        self.definition = definition
        self.policy_name = "syncr_policy_{0}".format(self.name.replace("/", "__"))

        self.trust = {}
        self.distrust = {}
        self.permission = []

    def setup(self):
        """Raise errors if the definition doesn't make sense"""
        self.description = self.definition.get("description", "No description provided!")
        for key, store in (("allow_to_assume_me", self.trust), ("disallow_to_assume_me", self.distrust)):
            for principal in listified(self.definition, key):
                for namespace, principal in self.make_trust_principal(principal).items():
                    if namespace not in store:
                        store[namespace] = set()

                    if not isinstance(principal, list):
                        principal = [principal]

                    for specified in principal:
                        store[namespace].add(specified)

            for namespace, vals in store.items():
                store[namespace] = sorted(list(vals))

        for key, default_allow in (("permission", None), ("allow_permission", True), ("deny_permission", False)):
            for policy in listified(self.definition, key):
                for statement in self.make_permission_statements(policy, allow=default_allow):
                    self.permission.append(statement)

    def resolve(self):
        """Make sure this user exists and has only what policies we want it to have"""
        # Get the permission and trust document
        # Make sure they're both valid before continuing
        trust_document = self.make_trust_document(self.trust, self.distrust)
        permission_document = self.make_permission_document(self.permission)

        role_info = self.amazon.role_info(self.name)
        if not role_info:
            self.amazon.create_role(self.name, trust_document, policies={self.policy_name: permission_document})
        else:
            self.amazon.modify_role(role_info, self.name, trust_document, policies={self.policy_name: permission_document})

        if self.definition.get("make_instance_profile"):
            self.amazon.make_instance_profile(self.name)

    def make_trust_principal(self, principal):
        """Make a trust statement"""
        result = dict((key, val) for key, val in principal.items() if key[0].isupper())

        for specified in listified(principal, "service"):
            if specified == "ec2":
                specified = "ec2.amazonaws.com"
            listify(result, "Service").append(specified)

        for specified in listified(principal, "federated"):
            listify(result, "Federated").append(specified)

        if "iam" in principal:
            listify(result, "AWS").extend(self.iam_arns_from_specification(principal))

        return result

    def make_permission_statements(self, policy, allow=None):
        """
        Make zero or more permissions statements from a policy definition

        If allow is None and no allow or Effect is specified, an error is raised
        Otherwise we assume the truthiness of allow
        """
        result = dict((key, val) for key, val in policy.items() if key[0].isupper())

        if allow is None and "Effect" not in policy:
            if not isinstance(policy.get("allow"), bool):
                raise BadPolicy("Need to specify whether we allow this policy or not", policy=policy, role=self.name)
            result["Effect"] = ["Deny", "Allow"][policy["allow"]]
        elif allow is not None:
            result["Effect"] = ["Deny", "Allow"][allow]

        for key, dest in (("action", "Action"), ("notaction", "NotAction")):
            for specified in listified(policy, key):
                listify(result, dest).append(specified)

        for key, dest in (("resource", "Resource"), ("notresource", "NotResource")):
            for specified in listified(policy, key):
                listify(result, dest).extend(self.fill_out_resources(specified))

        if "Resource" not in result and "NotResource" not in result:
            raise BadPolicy("No Resource or NotResource was defined for policy", role=self.name, policy=policy)

        if "Action" not in result and "NotAction" not in result:
            raise BadPolicy("No Action or NotAction defined for this policy", role=self.name, policy=policy)

        yield result

    def fill_out_resources(self, resources):
        """Fill out the resources"""
        for resource in as_list(resources):
            if isinstance(resource, basestring):
                yield resource
            elif isinstance(resource, dict):
                for resource in self.expand_resource(resource):
                    yield resource
            else:
                raise BadPolicy("Resource should be a string or a dictionary", resource=resource)

    def expand_resource(self, resource):
        """Return a resource string for given resource"""
        if "iam" in resource:
            for found in self.iam_arns_from_specification(resource):
                yield found

        elif "s3" in resource:
            for bucket_key in listify(resource, "s3"):
                yield "arn:aws:s3:::{0}".format(bucket_key)
                if '/' not in bucket_key:
                    yield "arn:aws:s3:::{0}/*".format(bucket_key)

        else:
            raise BadPolicy("Unknown resource type", resource=resource)

    def iam_arns_from_specification(self, specification):
        """Get us an iam arn from this specification"""
        provided_account = specification.get("account", "")
        if provided_account:
            if provided_account not in self.amazon.accounts:
                raise BadPolicy("Unknown account specified", account=provided_account, specification=specification)
            else:
                account_id = self.amazon.accounts[provided_account]
        else:
            account_id = self.amazon.account_id

        for name in listified(specification, "iam"):
            if name == "__self__":
                account_id = self.amazon.account_id
                name = "role/{0}".format(self.name)

            yield "arn:aws:iam::{0}:{1}".format(account_id, name)

    def make_trust_document(self, trust, distrust):
        """Make a document for trust or None if no trust or distrust"""
        if not trust and not distrust:
            return

        statement = {"Sid": "", "Action": "sts:AssumeRole", "Effect": "Allow"}

        if trust:
            statement["Principal"] = trust

        if distrust:
            statement["NotPrincipal"] = distrust

        return self.make_document([statement])

    def make_permission_document(self, permissions):
        """Return a document for these permissions, or None if no permissiosn"""
        if not permissions:
            return
        return self.make_document(permissions)

    def make_document(self, statements):
        """Make sure our document is valid and return it formatted correctly"""
        if not isinstance(statements, list):
            raise Exception("Statements should be a list!: got {0}".format(statements))

        document = {"Version": "2012-10-17", "Statement": statements}

        try:
            return json.dumps(document, indent=2)
        except (TypeError, ValueError) as err:
            raise InvalidDocument("Document wasn't valid json", role=self.name, error=err)




from iam_syncr.errors import SyncrError, InvalidConfiguration, ConflictingConfiguration, BadConfiguration, DuplicateItem
from iam_syncr.roles import Role, RoleRemoval

from collections import defaultdict
import logging

log = logging.getLogger("iam_syncr.syncr")

class Sync(object):
    """Knows how to interpret configuration for syncing"""

    def __init__(self, amazon):
        self.amazon = amazon

        self.types = {}
        self.configurations = defaultdict(list)

    def sync(self, combined):
        """Let's do this!"""
        for name in self.types:
            if name in combined:
                things = self.create_things(combined[name], name)
                self.setup_and_resolve(things)

    def register_default_types(self):
        """Register the default things syncr looks for"""
        self.register_type("roles", dict, Role, key_conflicts_with=["remove_roles"])
        self.register_type("remove_roles", list, RoleRemoval, key_conflicts_with=["roles"])

    def register_type(self, name, typ, kls, key_conflicts_with=None):
        """
        Register a type to be synced

        Override existing types
        """
        if key_conflicts_with and not isinstance(key_conflicts_with, list):
            key_conflicts_with = [key_conflicts_with]
        self.types[name] = (typ, key_conflicts_with, kls)

    def create_things(self, things, name):
        """Creates a list of Role objects from a roles_doc"""
        typ, _, kls = self.types[name]
        if typ is list:
            return [kls(thing, self.amazon) for thing in things]
        else:
            return [kls(thing, val, self.amazon) for thing, val in things.items()]

    def setup_and_resolve(self, things):
        """Runs setup on all the provided things and once they are setup, resolve them"""
        for thing in things:
            thing.setup()

        for thing in things:
            thing.resolve()

    def add(self, configuration, location, only_consider=None):
        """Add a new configuration"""
        if not self.types:
            raise SyncrError("Syncr doesn't know about anything, try syncr.register_default_types() first")

        if not isinstance(configuration, dict):
            raise InvalidConfiguration("Configuration needs to be a dict", found=type(configuration))

        for name in self.types:
            if not only_consider or name in only_consider:
                if name in configuration:
                    self.configurations[name].append((location, configuration[name]))

    def combine_configurations(self):
        """
        Return the combination of current configuration

        Or raise BadConfiguration(errors=[<error>, ...])
        """
        errors = []
        combined = {}
        for name, (typ, _, _) in self.types.items():
            for (location, configuration) in self.configurations[name]:
                nxt_errors = self.add_to_combined(combined, name, typ, configuration, location)
                if nxt_errors:
                    errors.extend(nxt_errors)

        conflicting = self.find_conflicting(combined)
        if conflicting:
            errors.extend(conflicting)

        if errors:
            raise BadConfiguration(errors=errors)

        return self.merge_combined(combined)

    def merge_combined(self, combined):
        """Merge our dictionary of {key: {thing: [(location, val), ...]}}"""
        merged = {}
        for key, things in combined.items():
            if things:
                make = self.types[key][0]
                if key not in merged:
                    merged[key] = make()

                for thing, found in things.items():
                    if found:
                        if make is list:
                            merged[key].append(thing)
                        else:
                            merged[key][thing] = found[0][1]

        return merged

    def find_conflicting(self, combined):
        """Return array of ConflictingConfiguration errors for any conflicting values"""
        errors = []
        complained_about = set()
        for key, collections in combined.items():
            for name, found in collections.items():
                if any(len(info) > 1 for info in found) and len(found) > 1:
                    errors.append(DuplicateItem(key=key, name=name, found=[info[0] for info in found]))

        for name, (_, conflicts_with, _) in sorted(self.types.items()):
            if conflicts_with and name in combined and any(conflictor in combined for conflictor in conflicts_with):
                for thing in combined[name]:
                    conflicting = [conflictor for conflictor in conflicts_with if thing in combined.get(conflictor, {})]
                    if any(conflicting):
                        identities = ["{0}{1}".format(category, thing) for category in [name] + conflicting]
                        if all(identity not in complained_about for identity in identities):
                            location_to_keys = defaultdict(set)
                            for key in [name] + conflicting:
                                for values in combined[key].values():
                                    for vals in values:
                                        location_to_keys[vals[0]].add(key)

                            found_in = "; ".join(sorted("{0}({1})".format(location, ', '.join(sorted(list(keys)))) for location, keys in location_to_keys.items()))
                            errors.append(ConflictingConfiguration("Found item in conflicting specifications", conflicting=thing, found_in=found_in))
                            for identity in identities:
                                complained_about.add(identity)

        return errors

    def add_to_combined(self, combined, key, expected_type, configuration, location):
        """
        Add the things under this key in the configuration to combined
        And return any errors that are found

        Complain if the key under the configuration is not the expected type

        Make sure to complain about duplicates in keys that are meant to be lists
        Record lists as {<thing>: [<location>, ....]} in combined.

        Record dictionaries in combined as {<thing>: [<location>, <configuration>]}
        """
        errors = []
        if not isinstance(configuration, expected_type):
            errors.append(InvalidConfiguration("Expected configuration of a different type", key=key, expected_type=expected_type, found=type(configuration), location=location))
        else:
            if key not in combined:
                combined[key] = defaultdict(list)

            if expected_type == list:
                track = set()
                duplicates = set()
                for thing in configuration:
                    if thing in track:
                        duplicates.add(thing)
                    else:
                        track.add(thing)

                if duplicates:
                    errors.append(InvalidConfiguration("Found duplicates in a list", key=key, location=location, duplicates=duplicates))
                else:
                    for thing in configuration:
                        combined[key][thing].append((location, ))
            else:
                for thing, val in configuration.items():
                    combined[key][thing].append((location, val))

        return errors




from setuptools import setup, find_packages
from iam_syncr import VERSION

setup(
      name = "iam_syncr"
    , version = VERSION
    , packages = ['iam_syncr'] + ['iam_syncr.%s' % pkg for pkg in find_packages('iam_syncr')]
    , include_package_data = True

    , install_requires =
      [ "rainbow_logging_handler"
      , "pyYaml"
      , "boto"
      ]

    , extras_require =
      { "tests":
        [ "noseOfYeti>=1.5.1"
        , "nose"
        , "mock"
        ]
      }

    , entry_points =
      { 'console_scripts' :
        [ 'iam_syncr = iam_syncr.executor:main'
        ]
      }

    # metadata for upload to PyPI
    , author = "Stephen Moore"
    , author_email = "stephen@rea-group.com"
    , description = "Syncs iam roles"
    , license = "MIT"
    , keywords = "iam amazon credentials"
    )



class SyncrError(Exception):
    """Helpful class for creating custom exceptions"""
    desc = ""

    def __init__(self, message="", **kwargs):
        self.kwargs = kwargs
        self.message = message
        super(SyncrError, self).__init__(message)

    def __str__(self):
        desc = self.desc
        message = self.message

        info = ["{0}={1}".format(k, v) for k, v in sorted(self.kwargs.items())]
        info = '\t'.join(info)
        if info and (message or desc):
            info = "\t{0}".format(info)

        if desc:
            if message:
                message = ". {0}".format(message)
            return '"{0}{1}"{2}'.format(desc, message, info)
        else:
            if message:
                return '"{0}"{1}'.format(message, info)
            else:
                return "{0}".format(info)

    def __eq__(self, error):
        """Say whether this error is like the other error"""
        return error.__class__ == self.__class__ and error.message == self.message and error.kwargs == error.kwargs

class InvalidDocument(SyncrError):
    desc = "Something wrong with this iam document"

class BadConfiguration(SyncrError):
    desc = "Something wrong with the configuration"

class InvalidConfiguration(SyncrError):
    desc = "Something wrong with this configuration"

class NoConfiguration(SyncrError):
    desc = "Didn't find any yaml files"

class ConflictingConfiguration(SyncrError):
    desc = "The configuration conflicts with itself"

class DuplicateItem(SyncrError):
    desc = "Item defined multiple times"

class BadRole(SyncrError):
    desc = "Bad definition of a role"

class BadPolicy(SyncrError):
    desc = "Bad definition of a policy"

class BadAmazon(SyncrError):
    desc = "Amazon said no"

